<?php

namespace Drupal\quickchat_sync\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Views;
use Drupal\quickchat_sync\QuickchatSyncApiClient;
use Drupal\Core\Config\ConfigFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Implements the module operation form.
 */
class OperationForm extends FormBase {

  /**
   * Maximum number of words for paragraph is 150.
   */
  const QUICKCHAT_MAX_NUMBERS_WORDS = 150;

  /**
   * Maximum number of characters for paragraph is 1000.
   */
  const QUICKCHAT_MAX_NUMBERS_CHARACTERS = 1000;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The quickchat_sync.api_client service.
   *
   * @var \Drupal\quickchat_sync\QuickchatSyncApiClient
   */
  protected $quickchatSyncApiClient;

  /**
   * {@inheritdoc}
   */
  public function __construct(ConfigFactoryInterface $config_factory, QuickchatSyncApiClient $quickchat_api_client) {
    $this->configFactory = $config_factory;
    $this->quickchatSyncApiClient = $quickchat_api_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('quickchat_sync.api_client')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'quickchat_sync.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'quickchat_sync_operation_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $model_name = NULL) {
    $config = $this->configFactory->get('quickchat_sync.settings');
    $sync = $config->get('sync');

    foreach ($sync as $model_id => $properties) {
      $machine_name = $properties['label_machine_name'];

      if ($machine_name === $model_name) {
        break;
      }
    }

    $form['#tree'] = TRUE;
    $form = [
      '#type' => 'fieldset',
      '#title' => $sync[$model_id]['label'],
      '#prefix' => '<div id="models-fieldset-wrapper">',
      '#suffix' => '</div>',
    ];

    $arguments = '';

    if (array_key_exists('view_arguments', $sync[$model_id])) {
      $arguments = $sync[$model_id]['view_arguments'];
    }

    $view_name = $sync[$model_id]['view_name'];
    $view_display = $sync[$model_id]['view_display'];
    $view_arguments = explode(',', $arguments);
    $preview = $this->getViewHtml($view_name, $view_display, $view_arguments);
    $preview = array_filter(explode('\n', $preview));
    $preview = $this->generateHtmlList($preview);

    $form[$model_id]['update'] = [
      '#type' => 'submit',
      '#value' => $this->t('Update model'),
      '#name' => $model_id . '_update',
      '#submit' => ['::updateCallback'],
      '#ajax' => [
        'callback' => '::ajaxSycCallback',
        'wrapper' => 'models-fieldset-wrapper',
      ],
    ];

    $form[$model_id]['rebuild'] = [
      '#type' => 'submit',
      '#value' => $this->t('Rebuild model'),
      '#name' => $model_id . '_rebuild',
      '#submit' => ['::rebuildCallback'],
      '#ajax' => [
        'callback' => '::ajaxSycCallback',
        'wrapper' => 'models-fieldset-wrapper',
      ],
    ];

    $form[$model_id]['preview'] = [
      '#type' => 'item',
      '#title' => $this->t('Preview'),
      '#markup' => $preview,
    ];

    $form[$model_id]['scenario_id'] = [
      '#type' => 'hidden',
      '#value' => $sync[$model_id]['scenario_id'],
    ];

    $form[$model_id]['token'] = [
      '#type' => 'hidden',
      '#value' => $sync[$model_id]['token'],
    ];

    $form[$model_id]['view_name'] = [
      '#type' => 'hidden',
      '#value' => $sync[$model_id]['view_name'],
    ];

    $form[$model_id]['view_display'] = [
      '#type' => 'hidden',
      '#value' => $sync[$model_id]['view_display'],
    ];

    $form[$model_id]['view_arguments'] = [
      '#type' => 'hidden',
      '#value' => $sync[$model_id]['view_arguments'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {}

  /**
   * Callback for both ajax-enabled buttons.
   *
   * Selects and returns the fieldset with the models sync in it.
   */
  public function ajaxSycCallback(array &$form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * Submit handler for the "rebuid" button.
   *
   * Rebuilds the corresponding line.
   */
  public function rebuildCallback(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $indexToRebuild = intval(strtok($trigger['#name'], '_rebuild'));
    $values = $form_state->getValues();
    $config = $this->configFactory->get('quickchat_sync.settings');
    $sync = $config->get('sync');

    $scenario_id = $form[$indexToRebuild]['scenario_id']['#value'];
    $token = $form[$indexToRebuild]['token']['#value'];

    $response = $this->executeOperation('retrain', $scenario_id, $token);

    if ($response) {
      if ($response->getStatusCode() == 200) {
        $this->messenger()->addMessage($this->t('The knowledge base has been rebuilt on Quickchat AI.'));
      }
      else {
        $this->messenger()->addError($this->t('@error', [
          '@error' => $response->getReasonPhrase(),
        ]));
      }
    }

    $form_state->setRebuild();
  }

  /**
   * Submit handler for the "update" button.
   *
   * Update the corresponding line.
   */
  public function updateCallback(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $indexToRebuild = intval(strtok($trigger['#name'], '_rebuild'));
    $values = $form_state->getValues();
    $config = $this->configFactory->get('quickchat_sync.settings');
    $sync = $config->get('sync');

    $scenario_id = $form[$indexToRebuild]['scenario_id']['#value'];
    $token = $form[$indexToRebuild]['token']['#value'];

    $view_name = $form[$indexToRebuild]['view_name']['#value'];
    $view_display = $form[$indexToRebuild]['view_display']['#value'];
    $view_arguments = [];

    $args = str_replace(' ', '', $form[$indexToRebuild]['view_arguments']['#value']);

    if ($args) {
      $view_arguments = explode(',', $args);
    }

    $html = $this->getViewHtml($view_name, $view_display, $view_arguments);

    $response = $this->executeOperation('update_knowledge_base', $scenario_id, $token, $html);

    if ($response) {
      $this->messenger()->deleteAll();
      if ($response->getStatusCode() == 200) {
        $this->messenger()->addMessage($this->t('The knowledge base has been updated on Quickchat AI.'));
      }
      else {
        $this->messenger()->addError($this->t('@error', [
          '@error' => $response->getReasonPhrase(),
        ]));
      }
    }

    $form_state->setRebuild();
  }

  /**
   * Helper function to get the render HTML from a given view.
   */
  private function getViewHtml($view_name, $view_display, $arguments = []) {
    $html = '';

    if ($view = Views::getView($view_name)) {
      $build = $view->buildRenderable($view_display, $arguments);
      $html = strip_tags(\Drupal::service('renderer')->renderPlain($build));
      $html = preg_replace('/\r|\n/', '', trim($html));
    }

    return rtrim($html);
  }

  /**
   * Helper to call the Quickchat API to train the given chatbot model.
   *
   * @param string $action
   *   Action to perform: fetch, retrain, update_name, update_short_description,
   *   update_knowledge_base, update_how_to_reach_you.
   * @param string $scenario_id
   *   ID associated with your Quickchat API implementation.
   *   update_knowledge_base, update_how_to_reach_you.
   * @param string $token
   *   Token to use for API limited access.
   * @param string $text
   *   Optional parameter, only to be provided for update actions.
   */
  private function executeOperation($action, $scenario_id, $token, $text = '') {
    return $this->quickchatSyncApiClient->drupal($action, $scenario_id, $token, $text);
  }

  /**
   * Helper to generate a HTML list using given array.
   *
   * @param array $values
   *   Array containing the values to parse into a HTML list.
   */
  private function generateHtmlList(array $values) {
    $return = '<ul>';

    foreach ($values as $value) {
      $errors = '';

      if (str_word_count($value) > OperationForm::QUICKCHAT_MAX_NUMBERS_WORDS) {
        $this->messenger()->addError($this->t('You entered @words words. Maximum number of words for paragraph is @max_words.', [
          '@words' => str_word_count($value),
          '@max_words' => OperationForm::QUICKCHAT_MAX_NUMBERS_WORDS,
        ]));
        $errors = '❌';
      }

      if (strlen($value) > OperationForm::QUICKCHAT_MAX_NUMBERS_CHARACTERS) {
        $this->messenger()->addError($this->t('You entered @characters characters. Maximum number of characters for paragraph is @max_characters.', [
          '@characters' => strlen($value),
          '@max_characters' => OperationForm::QUICKCHAT_MAX_NUMBERS_CHARACTERS,
        ]));
        $errors = '❌';
      }

      $return .= "<li>$errors $value</li>";
    }

    $return .= '</ul>';

    return $return;
  }

}
