<?php

namespace Drupal\Quickchat;

use Drupal\Core\Config\ConfigFactoryInterface;
use GuzzleHttp\ClientInterface;
use Drupal\Core\Url;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Exception\ServerException;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;

/**
 * Quickchat API client wrapper class.
 */
class QuickchatApiClient {

  /**
   * GuzzleHttp\ClientInterface definition.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * Quickchat API base url.
   *
   * @var string
   */
  protected $apiBaseUrl;

  /**
   * Quickchat API key.
   *
   * @var string
   */
  protected $apiKey;

  /**
   * Quickchat API scenario id.
   *
   * @var string
   */
  protected $apiScenarioId;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a new QuickchatApiClient object.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The Guzzle HTTP client.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The UUID service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   Logger service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(
    ClientInterface $http_client,
    ConfigFactoryInterface $config_factory,
    UuidInterface $uuid_service,
    MessengerInterface $messenger,
    LoggerChannelFactoryInterface $logger) {
    $this->httpClient = $http_client;
    $this->uuidService = $uuid_service;
    $config = $config_factory->get('quickchat.settings');
    $this->apiBaseUrl = $config->get('api_baseurl');
    $this->apiKey = $config->get('api_key');
    $this->apiScenarioId = $config->get('scenario_id');
    $this->messenger = $messenger;
    $this->logger = $logger->get('quickchat');
  }

  /**
   * Performs an Quickchat API request. Wraps the Guzzle HTTP client.
   *
   * @param string $method
   *   HTTP method.
   * @param string $endpoint
   *   The API endpoint to call.
   * @param array $body
   *   The body to send to the API endpoint.
   * @param array $query
   *   API call query parameters.
   *
   * @return \Psr\Http\Message\ResponseInterface
   *   The request response.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   *
   * @see https://www.quickchat.ai/docs/#introduction
   */
  public function request($method,
    $endpoint,
    array $body = [],
    array $query = []) {

    $endpoint = $this->apiBaseUrl . $endpoint;

    if ($method === 'GET') {
      $endpoint = Url::fromUri($endpoint, [
        'query' => $query,
      ])->toString();
    }

    $request = [
      'headers' => [
        'Content-Type' => 'application/json',
      ],
      'timeout' => 20,
    ];

    if ($body) {
      $request['body'] = Json::encode($body);
    }

    try {
      $response = $this->httpClient->request($method, $endpoint, $request);
      if ($response->getStatusCode() != 200) {
        $this->messenger->addError(t('@error', [
          '@error' => $response->getReasonPhrase(),
        ]));
        $this->logger->error('Request failed: @response.', [
          '@response' => $response->getReasonPhrase(),
        ]);
      }

      $this->logger->notice('Response from %endpoint: 200 @response.', [
        '%endpoint' => $endpoint,
        '@response' => $response->getReasonPhrase(),
      ]);
    }
    catch (RequestException | ServerException | ClientException | ConnectException $e) {
      $this->logger->error(t('@error', [
        '@error' => $e->getMessage(),
      ]));

      $this->messenger->addError(t('@error', [
        '@error' => $e->getMessage(),
      ]));
    }

    return $response;
  }

  /**
   * Initialise new conversation / new user history.
   *
   * This endpoint performs a single conversation exchange.
   *
   * @param string $text
   *   User input message.
   * @param string $scenario_id
   *   ID associated with your Quickchat API implementation.
   * @param string $conv_id
   *   This is an optional parameter. Request which does not contain conv_id
   *   starts a new conversation and newly-generated conv_id is returned.
   *
   * @return array
   *   Json from API response parsed to array format.
   *
   * @see https://www.quickchat.ai/docs/#initialise-new-conversation--new-user-history
   */
  public function chat($text, $scenario_id = NULL, $conv_id = NULL) {
    $data = [
      'api_key' => $this->apiKey,
      'text' => $text,
    ];

    if ($conv_id) {
      $data['conv_id'] = $conv_id;
    }

    if ($scenario_id) {
      $data['scenario_id'] = $scenario_id;
    }
    else {
      $data['scenario_id'] = $this->apiScenarioId;
    }

    $response = $this->request('POST', '/chat', $data);

    return $response;
  }

  /**
   * Send message to AI. Say something to the AI and receive a response.
   *
   * @param string $text
   *   User input message.
   * @param string $scenario_id
   *   ID associated with your Quickchat API implementation.
   * @param bool $skip_reply
   *   If true, AI does not provide reply.
   *   Message from user appended to conversation history.
   * @param bool $override_ai
   *   If true, message from AI appended to conversation history.
   * @param string $conv_id
   *   This is an optional parameter. Request which does not contain conv_id
   *   starts a new conversation and newly-generated conv_id is returned.
   *
   * @return array
   *   Json from API response parsed to array format.
   *
   * @see https://www.quickchat.ai/docs/#send-message-to-ai
   */
  public function chatAi($text, $scenario_id, $skip_reply, $override_ai, $conv_id = NULL) {
    $data = [
      'api_key' => $this->apiKey,
      'text' => $text,
      'skip_reply' => $skip_reply,
      'override_ai' => $override_ai,
    ];

    if ($conv_id) {
      $data['conv_id'] = $conv_id;
    }

    if ($scenario_id) {
      $data['scenario_id'] = $scenario_id;
    }
    else {
      $data['scenario_id'] = $this->apiScenarioId;
    }

    return $this->request('POST', '/chat', $data);
  }

}
